import textwrap
from datetime import datetime
from enum import Enum
from typing import Dict, NamedTuple, Optional, List, Iterator

from re import sub

from grid.openapi.rest import ApiException
from typing import Dict, NamedTuple, Optional, List, Iterator
from re import sub

# use simple import to prevent cyclical dependency with artifacts
import grid.sdk.artifacts
from grid.sdk.datastores import Datastore
from grid.sdk.run_components import Actions, Resources, ScratchSpace
from grid.sdk import env
from grid.sdk.rest.client import GridRestClient
from grid.sdk.affirmations import affirm, is_not_deleted, is_created, is_not_shallow
from grid.sdk.rest.experiments import get_experiment_from_id, get_experiment_by_name, update_experiment, \
    delete_experiment, list_artifacts
from grid.sdk.user import User
from grid.sdk.client import create_swagger_client

from grid.openapi.models import (
    V1ExperimentState, V1PackageManager, Externalv1Experiment, V1ListExperimentArtifactsResponse
)
from grid.sdk.utilities import SPECIAL_NAME_TO_SKIP_OBJECT_INIT


class ExperimentName(NamedTuple):
    name: str
    username: Optional[str]

    @classmethod
    def from_external_format(cls, name: str):
        """User can pass experiments as username:experiment_name to fetch other users experiments.
        """
        username = None
        split = name.split(":")
        if len(split) > 2:
            raise ValueError(f"Error while parsing {name}. Use the format <username>:<experiment-name>")
        elif len(split) == 2:
            username = split[0]
            name = split[1]

        return cls(name=name, username=username)


class Experiment:
    _client: GridRestClient

    _created_at: datetime
    _description: str
    _id: str
    _name: str
    _updated_at: datetime
    _actions: Actions
    _cluster_id: str
    _command: List[str]
    _datastore: Datastore
    _datastore_mount_dir: str
    _desired_state: 'ExperimentState'
    _dry_run: bool
    _env: Dict[str, str]
    _runner_framework: str
    _runner_dockerfile: str
    _runner_requirements_file: str
    _runner_package_manager: str
    _instance_type: str
    _resources: Resources
    _run_controller_id: str
    _scratch: ScratchSpace
    _source_code: str
    _use_spot: bool
    _user_id: str
    _cost: float
    _deleted_timestamp: datetime
    _finish_timestamp: datetime
    _hourly_cost: float
    _message: str
    _status: 'ExperimentState'
    _start_timestamp: datetime
    _user: User

    _is_deleted: bool
    _is_created: bool
    _is_shallow: bool

    def __init__(
        self,
        name: str,
        cluster_id: Optional[str] = None,
    ):
        # --------------------------------------------------------------------------- #
        #    This should be the first block that goes into the constructor of the     #
        #    resource object. This block sets the correct values for the private      #
        #    attributes which is then later picked up by the decorator(s) to take     #
        #    right actions. It also initialize the _client object and cluster_id      #
        #    which is required by the downstream methods regardless of the object     #
        #    is completely initialized or not. Most importantly, this blocks checks   #
        #    for the name argument to decide if it's a call from other internal       #
        #    methods to create a shallow object. This is done by checking the         #
        #    special name variable. Other methods that already has the backend        #
        #    response fetched, can use this to create the object without the backend  #
        #    call and then fill-in the response they already have.                    #
        #                                                                             #
        self._client = GridRestClient(api_client=create_swagger_client())
        cluster_id = cluster_id or env.CONTEXT
        self._is_shallow = False
        self._cluster_id = cluster_id
        if name == SPECIAL_NAME_TO_SKIP_OBJECT_INIT:
            self._is_shallow = True
            self._is_created = False
            self._is_deleted = False
            return
        #                                                                             #
        # --------------------------------------------------------------------------- #
        try:
            exp = get_experiment_by_name(client=self._client, cluster_id=self._cluster_id, experiment_name=name)
            self._setup_from_response(exp)
            return
        except KeyError:
            raise RuntimeError(
                f"Experiment {name} does not exist. If you are here "
                f"for creating experiment, use Run(...) since experiments "
                f"are grouped under Run object and can only be created "
                f"using the Run abstraction"
            ) from None

    @property
    @affirm(is_not_shallow)
    def exists(self) -> bool:
        return self._is_created

    @classmethod
    def _from_id(cls, exp_id: str, cluster_id: Optional[str] = env.CONTEXT) -> "Experiment":
        instance = cls(name=SPECIAL_NAME_TO_SKIP_OBJECT_INIT, cluster_id=cluster_id)
        instance._id = exp_id
        return instance

    def _setup_from_response(self, exp: 'Externalv1Experiment'):
        self._is_deleted = exp.status.phase == V1ExperimentState.DELETED
        self._is_created = True
        self._is_shallow = False

        spec = exp.spec
        status = exp.status
        image = spec.image

        # TODO - setup datastore
        self._name = exp.name
        self._id = exp.id
        self._created_at = exp.created_at
        self._description = exp.description
        self._updated_at = exp.updated_at

        # spec attributes
        self._cluster_id = spec.cluster_id
        self._command = spec.command
        self._desired_state = ExperimentState.from_api_spec(spec.desired_state)
        self._dry_run = spec.dry_run
        self._env = spec.env
        self._instance_type = spec.instance_type
        self._run_controller_id = spec.run_controller_id
        self._source_code = spec.source_code
        self._use_spot = spec.use_spot
        self._user_id = spec.user_id
        # TODO - set actions, resources and scratch

        # image attributes
        self._runner_framework = image.framework
        self._runner_dockerfile = image.dockerfile
        # TODO - understand why these values could be empty from the BE
        self._runner_requirements_file = image.dependency_file_info.path if image.dependency_file_info else None
        self._runner_package_manager = image.dependency_file_info.package_manager if image.dependency_file_info else None

        # status attributes
        self._cost = status.cost
        self._deleted_timestamp = status.deleted_timestamp
        self._finish_timestamp = status.finish_timestamp
        self._hourly_cost = status.hourly_cost
        self._message = status.message
        self._status = ExperimentState.from_api_spec(status.phase)
        self._start_timestamp = status.start_timestamp

        if not hasattr(self, '_user') or self._user is None:
            # TODO - fetch user from ID - we don't have that in BE
            self._user = User(user_id=spec.user_id, username="", first_name="", last_name="")

    @property
    def id(self) -> str:
        return self._id

    @affirm(is_not_shallow, is_not_deleted)
    def __repr__(self):
        return textwrap.dedent(
            f"""\
            {self.__class__.__name__}(
                {"name": <18} = {self._name},
                {"desired_state": <18} = {self.desired_state.value},
                {"status": <18} = {self.status.value},
                {"source_code": <18} = {self._source_code},
                {"command": <18} = {self._command},
                {"user": <18} = {self.user.user_id},
                {"cluster_id": <18} = {self._cluster_id},
                {"created_at": <18} = {self._created_at},
                {"start_timestamp": <18} = {self._start_timestamp},
                {"finished_timestamp": <18} = {self._finish_timestamp},
            )"""
        )

    def _unshallow(self):
        """ If the object is a shallow (i.e. only has an id and `_is_shallow` attribute is True)
        object, this method can be triggered to get the full object from the BE. It is designed
        to be called only from the `is_not_shallow` decorator and should not be called directly.
        """
        if not self._is_shallow:
            raise RuntimeError('Experiment is already unshallow')
        if not hasattr(self, '_id') or self._id is None:
            raise RuntimeError("Cannot unshallow resource without a valid Experiment id")
        self._is_shallow = False
        try:
            exp = get_experiment_from_id(self._client, experiment_id=self._id, cluster_id=self._cluster_id)
        except ApiException as e:  # TODO change to GridException
            if hasattr(e, 'reason') and e.reason == "Not Found":
                self._is_deleted = True
                self._status = ExperimentState.DELETED
        else:
            self._setup_from_response(exp)

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def created_at(self) -> datetime:
        """created_at of the experiment configured as part of the run.
        """
        return self._created_at

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def description(self) -> str:
        """description of the experiment configured as part of the run.
        """
        return self._description

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def name(self) -> str:
        """name of the experiment configured as part of the run.
        """
        return self._name

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def actions(self) -> Actions:
        """actions of the experiment configured as part of the run.
        """
        return self._actions

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def cluster_id(self) -> str:
        """cluster_id of the experiment configured as part of the run.
        """
        return self._cluster_id

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def command(self) -> List[str]:
        """command of the experiment configured as part of the run.
        """
        return self._command

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def datastore(self) -> Datastore:
        """datastores of the experiment configured as part of the run.
        """
        return self._datastore

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def datastore_mount_dir(self) -> str:
        return self._datastore_mount_dir

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def desired_state(self) -> 'ExperimentState':
        """desired_state of the experiment configured as part of the run.
        """
        return self._desired_state

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def dry_run(self) -> bool:
        """dry_run of the experiment configured as part of the run.
        """
        return self._dry_run

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def env(self) -> Dict[str, str]:
        """env of the experiment configured as part of the run.
        """
        return self._env

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def runner_framework(self) -> str:  # from _image: 'Externalv1ImageSpec'
        """runner_framework of the experiment configured as part of the run.
        """
        return self._runner_framework

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def runner_dockerfile(self) -> str:  # from _image: 'Externalv1ImageSpec'
        """runner_dockerfile of the experiment configured as part of the run.
        """
        return self._runner_dockerfile

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def runner_requirements_file(self) -> str:  # from _image: 'Externalv1ImageSpec'
        """runner_requirements_file of the experiment configured as part of the run.
        """
        return self._runner_requirements_file

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def runner_package_manager(self) -> 'V1PackageManager':  # from _image: 'Externalv1ImageSpec'
        """runner_package_manager of the experiment configured as part of the run.
        """
        return self._runner_package_manager

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def instance_type(self) -> str:
        """instance_type of the experiment configured as part of the run.
        """
        return self._instance_type

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def resources(self) -> Resources:
        """resources of the experiment configured as part of the run.
        """
        return self._resources

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def run_controller_id(self) -> str:
        """run_controller_id of the experiment configured as part of the run.
        """
        return self._run_controller_id

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def scratch(self) -> ScratchSpace:
        """scratch of the experiment configured as part of the run.
        """
        return self._scratch

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def source_code(self) -> str:
        """source_code of the experiment configured as part of the run.
        """
        return self._source_code

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def use_spot(self) -> bool:
        """use_spot of the experiment configured as part of the run.
        """
        return self._use_spot

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def user(self) -> User:
        """user who owns the of the experiment/run.
        """
        return self._user

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def cost(self) -> float:
        """cost of the experiment configured as part of the run.
        """
        return self._cost

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def deleted_timestamp(self) -> datetime:
        """deleted_timestamp of the experiment configured as part of the run.
        """
        return self._deleted_timestamp

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def finish_timestamp(self) -> datetime:
        """finish_timestamp of the experiment configured as part of the run.
        """
        return self._finish_timestamp

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def hourly_cost(self) -> float:
        """hourly_cost of the experiment configured as part of the run.
        """
        return self._hourly_cost

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def message(self) -> str:
        """message of the experiment configured as part of the run.
        """
        return self._message

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def status(self) -> 'ExperimentState':
        """status of the experiment configured as part of the run.
        """
        return self._status

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def start_timestamp(self) -> datetime:
        """start_timestamp of the experiment configured as part of the run.
        """
        return self._start_timestamp

    @affirm(is_not_shallow, is_not_deleted, is_created)
    def cancel(self) -> bool:
        """Request the grid platform cancels the experiment
        """
        # do not perform POST if we know that we don't have to
        exp = update_experiment(
            client=self._client,
            cluster_id=self._cluster_id,
            experiment_id=self._id,
            desired_state=V1ExperimentState.CANCELLED
        )
        self._setup_from_response(exp)
        return True

    @affirm(is_not_shallow, is_not_deleted, is_created)
    def delete(self) -> bool:
        """Request the grid platform cancels the experiment
        """
        # do not perform POST if we know that we don't have to
        delete_experiment(client=self._client, cluster_id=self._cluster_id, experiment_id=self._id)
        self._is_deleted = True
        self._is_created = False
        return True

    @property
    def run_name(self):
        """Removes the experiment identifier to get the run name
        """
        return sub(r'-exp[0-9]+$', '', self.name)

    @property
    @affirm(is_not_shallow, is_not_deleted, is_created)
    def artifacts(self) -> Iterator['grid.sdk.artifacts.Artifact']:
        """Request a list of artifacts created by the experiment.

        Yields
        -------
        Iterator[Artifact]
            Each artifact contains its filename and URL for downloading.
        """

        # used to get artifacts from a response
        def get_artifacts(resp: V1ListExperimentArtifactsResponse) -> Iterator['grid.sdk.artifacts.Artifact']:
            for artifact in resp.artifacts:
                yield grid.sdk.artifacts.Artifact(
                    url=artifact.url,
                    filename=artifact.filename,
                    experiment_name=self.name,
                    run_name=self.run_name,
                )

        # get the first page of artifacts
        page_token = ""
        page_size = "25"
        resp = list_artifacts(self._client, self.cluster_id, self._id, page_token, page_size)
        for artifact in get_artifacts(resp):
            yield artifact

        # get the rest
        while resp.next_page_token != "":
            resp = list_artifacts(self._client, self.cluster_id, self._id, resp.next_page_token, page_size)
            for artifact in get_artifacts(resp):
                yield artifact


class ExperimentState(Enum):
    """
    The state of an experiment.
    """
    UNSPECIFIED = "unspecified"
    PENDING = "pending"
    QUEUED = "queued"
    RUNNING = "running"
    SUCCEEDED = "succeeded"
    FAILED = "failed"
    CANCELLED = "cancelled"
    DELETED = "deleted"
    PAUSED = "paused"
    IMAGE_BUILDING = "image_building"

    @classmethod
    def from_api_spec(cls, status: V1ExperimentState) -> 'ExperimentState':
        parsed = str(status).lower().split('_', maxsplit=2)[-1]
        return cls(parsed)



from pathlib import Path
from typing import Dict, Optional, List, TYPE_CHECKING

import validators

from grid.openapi import (
    Externalv1Datastore,
    Externalv1DatastoreSpec,
    V1GetDatastoreResponse,
    V1DatastoreSourceType,
    V1CreateDatastoreRequest,
    V1CreateDatastoreResponse,
    V1DatastoreUploadObject,
    Body2,
    V1PresignedUploadUrl,
    Body3,
    V1CompletePresignedUrlsUploadObject,
    V1CompletePresignedUrlUpload,
    V1DatastoreFSxSpec,
)
from grid.sdk.rest.client import GridRestClient
from grid.sdk.rest.exceptions import throw_with_message
from grid.sdk.user_messaging import errors

if TYPE_CHECKING:
    from grid.sdk.utils.datastore_uploads import DataFile


@throw_with_message
def create_datastore(
    c: GridRestClient,
    cluster_id: str,
    name: str,
    source: str,
    no_copy_data_source: bool,
    version: int = 0,
    fsx_enabled: bool = False,
    fsx_capacity_gib: int = 1200,
    fsx_throughput_mbs_tib: int = 125,
    fsx_preloading: bool = False
) -> V1CreateDatastoreResponse:
    """Create a new datastore.

    Parameters
    ----------
    c
        Client
    cluster_id
        Cluster ID to use for uploading files.
    name
        Human readable name assigned to the datastore.
    source
        the source string as passed in by the user on the CLI or SDK.
    no_copy_data_source
        do not copy bucket data, create a shallow clone.
    version
        The version of the datastore to create with some same `name`.
    fsx_enabled
        Create an FSx backed datastore instead of a conventional one.
    fsx_capacity_gib
        The capacity of the FSx file system backing a datastore in GiB, only to be used with fsx_source_type=True
    fsx_throughput_mbs_tib
        The throughput per unit of storage for the datastore in MB/s/TiB, only to be used with fsx_source_type=True
    fsx_preloading
        Whether to preload the data on datastore creation, only to be used with fsx_source_type=True
    """
    if Path(source).exists():
        source_type = V1DatastoreSourceType.EXPANDED_FILES
        if no_copy_data_source is True:
            raise ValueError(errors.datastore_invalid_no_copy_option(source))
        elif fsx_enabled is True:
            raise ValueError(errors.datastore_invalid_fsx_option(source))

    elif validators.url(source) is True:
        source_type = V1DatastoreSourceType.HTTP_URL
        if no_copy_data_source is True:
            raise ValueError(errors.datastore_invalid_no_copy_option(source))
        elif fsx_enabled is True:
            raise ValueError(errors.datastore_invalid_fsx_option(source))

    elif source.startswith("s3://"):
        if not source.endswith("/"):
            raise (ValueError(errors.datastore_s3_source_does_not_end_in_slash(source)))
        if no_copy_data_source is True:
            source_type = V1DatastoreSourceType.OBJECT_STORE_REFERENCE_ONLY
        elif fsx_enabled:
            source_type = V1DatastoreSourceType.FSX
        else:
            source_type = V1DatastoreSourceType.OBJECT_STORE

    else:
        raise ValueError(errors.datastore_invalid_source(source))

    if source_type == V1DatastoreSourceType.EXPANDED_FILES:
        source = ""  # expanded file source should not be set to a value in backend.

    fsx_spec = None

    if fsx_enabled is True:
        if fsx_throughput_mbs_tib not in [125, 250, 500, 1000]:
            raise ValueError(errors.datastore_invalid_fsx_throughput(fsx_throughput_mbs_tib))
        if fsx_capacity_gib != 1200 and fsx_capacity_gib % 2400 != 0:
            raise ValueError(errors.datastore_invalid_fsx_capacity(fsx_capacity_gib))

        fsx_spec = V1DatastoreFSxSpec(
            storage_capacity_gib=fsx_capacity_gib,
            storage_throughput_mb_s_tib=fsx_throughput_mbs_tib,
            preload_data_on_create=fsx_preloading,
        )

    resp = c.datastore_service_create_datastore(
        spec_cluster_id=cluster_id,
        body=V1CreateDatastoreRequest(
            name=name,
            spec=Externalv1DatastoreSpec(
                cluster_id=cluster_id,
                source=source,
                source_type=source_type,
                version=version,
                fsx_spec=fsx_spec,
            ),
        ),
    )
    return resp


def delete_datastore(c: 'GridRestClient', cluster_id: str, datastore_id: str) -> None:
    """Delete a datastore on a provided cluster with the given datastore id.
    """
    c.datastore_service_delete_datastore(cluster_id=cluster_id, id=datastore_id)


def datastore_upload_object_from_data_file(df: 'DataFile') -> V1DatastoreUploadObject:
    if not df.relative_file_name:
        raise ValueError(f"relative_file_name is a required field. Not found in df: {df}")
    if not df.part_count:
        raise ValueError(f"part_count is a required field. Not found in df: {df}")

    urls = None
    if len(df.tasks) > 0:
        urls = []
        for task in df.tasks:
            if not task.part_number:
                raise ValueError(f"part_number is required if datafile tasks are defined. df: {df} task: {task}")
            if not task.url:
                raise ValueError(f"url is required if datafile tasks are defined. df: {df} task: {task}")
            urls.append(V1PresignedUploadUrl(
                part_number=task.part_number,
                url=task.url,
            ))

    return V1DatastoreUploadObject(
        key=df.relative_file_name,
        part_count=str(df.part_count),
        upload_id=df.upload_id,
        expiry_time=df.expiry_time,
        urls=urls
    )


@throw_with_message
def create_presigned_urls(
    c: GridRestClient, cluster_id: str, datastore_id: str, upload_objects: List[V1DatastoreUploadObject]
) -> List[V1DatastoreUploadObject]:
    """Create presigned URLs for uploading files to a datastore.

    Parameters
    ----------
    c
        Client
    cluster_id
        Cluster ID to use for uploading files.
    datastore_id
        Datastore ID to use for uploading files.
    upload_objects
        List of objects to upload.
    """
    resp = c.datastore_service_create_datastore_presigned_urls(
        cluster_id=cluster_id,
        datastore_id=datastore_id,
        body=Body2(objects=upload_objects),
    )
    return resp.objects


@throw_with_message
def complete_presigned_url_upload(
    c: GridRestClient, cluster_id: str, datastore_id: str, data_files: List['DataFile']
) -> None:
    """Complete uploading files to a datastore.

    Parameters
    ----------
    c
        Client
    cluster_id
        Cluster ID to use for uploading files.
    datastore_id
        Datastore ID to use for uploading files.
    data_files
        List of data files to upload.
    """
    completed_objects = []
    for df in data_files:
        parts = []
        # Upload parts must always be sorted by part_number or an error will occur.
        # They should never be "unsorted" in this list, but this is a bit of defensive
        # programming to ensure that this is the case
        tasks = sorted(df.tasks, key=lambda x: int(x.part_number))
        for task in tasks:
            parts.append(V1CompletePresignedUrlUpload(
                part_number=task.part_number,
                etag=task.etag,
            ))

        completed_objects.append(
            V1CompletePresignedUrlsUploadObject(key=df.relative_file_name, upload_id=df.upload_id, urls=parts)
        )

    c.datastore_service_complete_datastore_presigned_urls_upload(
        cluster_id=cluster_id,
        datastore_id=datastore_id,
        body=Body3(objects=completed_objects),
    )


@throw_with_message
def mark_datastore_upload_complete(c: GridRestClient, cluster_id: str, datastore_id: str):
    """Tell the grid backend that all datastore files have been uploaded.

    Failing to call this method will indefintely hold the datastore desired
    state as "uploading"; meaning that it will never be optimized or made
    available to compute services.

    Parameters
    ----------
    c
        Client
    cluster_id
        The cluster ID the datastore is uploaded to
    datastore_id
        The datastore ID
    """
    c.datastore_service_mark_datastore_upload_complete(cluster_id=cluster_id, datastore_id=datastore_id)


@throw_with_message
def datastore_id_from_name(c: GridRestClient, cluster_id: str, name: str, version: Optional[str] = None) -> str:
    """Find the id of a datastore with some name (and optionally version) on a cluster.

    Parameters
    ----------
    c
        Client
    cluster_id
        which cluster should be used to find the datastore in.
    name
        the name of the datastore to find the ID of
    version
        The version of the datastore with ``name`` to find the ID of.
        NOTE: If no ``version`` argument is present, then the maximum
        version of the datastore will be used.

    Returns
    -------
    str
       The ID of the datastore.
    """
    dstores = c.datastore_service_list_datastores(cluster_id=cluster_id, available=True)

    datastore_versions: Dict[int, Externalv1Datastore] = {}

    for dstore in dstores.datastores:
        if dstore.name == name:
            spec = dstore.spec
            datastore_versions[spec.version] = dstore

    if version is None:
        # use the max version available
        version = max(datastore_versions.keys())

    try:
        return datastore_versions[version].id
    except KeyError:
        raise KeyError(f"no datastore exists with name: {name}")


@throw_with_message
def get_datastore_from_id(c: GridRestClient, datastore_id: str, cluster_id: str) -> V1GetDatastoreResponse:
    dstore = c.datastore_service_get_datastore(cluster_id=cluster_id, id=datastore_id)
    return dstore


@throw_with_message
def get_datastore_from_name(client: GridRestClient, cluster_id: str, datastore_name: str, version: int):
    ds_version_to_id_map = {}
    datastores = get_datastore_list(client=client, cluster_id=cluster_id)

    for datastore in datastores:
        if datastore.name == datastore_name:
            ds_version_to_id_map[datastore.spec.version] = datastore.id
            # return if the version is the same
            if version and datastore.spec.version == version:
                return datastore

    # if version is not specified, return the latest version
    if version and len(ds_version_to_id_map) > 0:
        max_version = max(ds_version_to_id_map.keys())
        return ds_version_to_id_map[max_version]

    raise KeyError(f'Datastore with name {datastore_name} and version {version} not found')


@throw_with_message
def get_datastore_list(client: GridRestClient,
                       cluster_id: str,
                       user_ids: Optional[List[str]] = None) -> List[Externalv1Datastore]:
    """get a list of the datastores the user has access to on the cluster

    Parameters
    ----------
    client
        the rest client to interact with the grid-backend from.
    cluster_id
        the cluster id to get a list of datastores for
    user_ids
        the user_ids to get a list of datastores for (authenticated client user
        must have authorization to list datastores of teammates). default=None
    """
    user_ids = user_ids or []
    resp = client.datastore_service_list_datastores(cluster_id=cluster_id, user_ids=user_ids, available=True)
    return resp.datastores


def datastore_dsn_from_id(id: str) -> str:
    """Return the DSN of the datastore

    Parameters
    ----------
    id
        datastore ID to convert into DSN.

    Returns
    -------
    str
        DSN form of the datastore ID.
    """
    return f"datastore://grid/{id}"


def datastore_id_from_dsn(dsn: str) -> str:
    """Return the id of a datastore from a DSN.

    Parameters
    ----------
    dsn
        DSN string to convert into an ID

    Returns
    -------
    str
        ID of the datastore DSN string.
    """
    # convert ``datastore://grid/{id}`` -> ``['datastore:', '', 'grid', '{id}']``
    # convert ``datastore://grid/{id}/`` -> ``['datastore:', '', 'grid', '{id}', '']``
    # ... (we want the last element)
    parts = dsn.split('/')
    if dsn.endswith('/'):
        if len(parts) < 2:
            raise RuntimeError(f"Internal Error. invalid datastore dsn format while parsing ID. dsn={dsn}")
        return parts[-2]
    return parts[-1]



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


from __future__ import absolute_import

import re  # noqa: F401
from typing import TYPE_CHECKING, Any

# python 2 and python 3 compatibility library
import six

from grid.openapi.api_client import ApiClient

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *


class DatastoreServiceApi(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    Ref: https://github.com/swagger-api/swagger-codegen
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

    def datastore_service_complete_datastore_presigned_urls_upload(self, cluster_id: 'str', datastore_id: 'str', body: 'Body3', **kwargs) -> 'V1CompleteDatastorePresignedUrlsUploadResponse':  # noqa: E501
        """CompleteDatastorePresignedUrlsUpload tells grid that all chunks of a file have been uploaded.  # noqa: E501

        This must be called once all parts of each presigned url requested for the datastore are processed on the client side. Failure to call this will make the uploaded file content be discarded from the datastore.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_complete_datastore_presigned_urls_upload(cluster_id, datastore_id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: the cluster the datastore is being created on. (required)
        :param str datastore_id: the id of the datastore. (required)
        :param Body3 body: (required)
        :return: V1CompleteDatastorePresignedUrlsUploadResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.datastore_service_complete_datastore_presigned_urls_upload_with_http_info(cluster_id, datastore_id, body, **kwargs)  # noqa: E501
        else:
            (data) = self.datastore_service_complete_datastore_presigned_urls_upload_with_http_info(cluster_id, datastore_id, body, **kwargs)  # noqa: E501
            return data

    def datastore_service_complete_datastore_presigned_urls_upload_with_http_info(self, cluster_id: 'str', datastore_id: 'str', body: 'Body3', **kwargs) -> 'V1CompleteDatastorePresignedUrlsUploadResponse':  # noqa: E501
        """CompleteDatastorePresignedUrlsUpload tells grid that all chunks of a file have been uploaded.  # noqa: E501

        This must be called once all parts of each presigned url requested for the datastore are processed on the client side. Failure to call this will make the uploaded file content be discarded from the datastore.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_complete_datastore_presigned_urls_upload_with_http_info(cluster_id, datastore_id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: the cluster the datastore is being created on. (required)
        :param str datastore_id: the id of the datastore. (required)
        :param Body3 body: (required)
        :return: V1CompleteDatastorePresignedUrlsUploadResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['cluster_id', 'datastore_id', 'body']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method datastore_service_complete_datastore_presigned_urls_upload" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'cluster_id' is set
        if self.api_client.client_side_validation and ('cluster_id' not in params or
                                                       params['cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `cluster_id` when calling `datastore_service_complete_datastore_presigned_urls_upload`")  # noqa: E501
        # verify the required parameter 'datastore_id' is set
        if self.api_client.client_side_validation and ('datastore_id' not in params or
                                                       params['datastore_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `datastore_id` when calling `datastore_service_complete_datastore_presigned_urls_upload`")  # noqa: E501
        # verify the required parameter 'body' is set
        if self.api_client.client_side_validation and ('body' not in params or
                                                       params['body'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `body` when calling `datastore_service_complete_datastore_presigned_urls_upload`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'cluster_id' in params:
            path_params['clusterId'] = params['cluster_id']  # noqa: E501
        if 'datastore_id' in params:
            path_params['datastoreId'] = params['datastore_id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{clusterId}/datastores/{datastoreId}/presigned-urls/completed', 'POST',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1CompleteDatastorePresignedUrlsUploadResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def datastore_service_create_datastore(self, spec_cluster_id: 'str', body: 'V1CreateDatastoreRequest', **kwargs) -> 'V1CreateDatastoreResponse':  # noqa: E501
        """datastore_service_create_datastore  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_create_datastore(spec_cluster_id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str spec_cluster_id: (required)
        :param V1CreateDatastoreRequest body: (required)
        :return: V1CreateDatastoreResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.datastore_service_create_datastore_with_http_info(spec_cluster_id, body, **kwargs)  # noqa: E501
        else:
            (data) = self.datastore_service_create_datastore_with_http_info(spec_cluster_id, body, **kwargs)  # noqa: E501
            return data

    def datastore_service_create_datastore_with_http_info(self, spec_cluster_id: 'str', body: 'V1CreateDatastoreRequest', **kwargs) -> 'V1CreateDatastoreResponse':  # noqa: E501
        """datastore_service_create_datastore  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_create_datastore_with_http_info(spec_cluster_id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str spec_cluster_id: (required)
        :param V1CreateDatastoreRequest body: (required)
        :return: V1CreateDatastoreResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['spec_cluster_id', 'body']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method datastore_service_create_datastore" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'spec_cluster_id' is set
        if self.api_client.client_side_validation and ('spec_cluster_id' not in params or
                                                       params['spec_cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `spec_cluster_id` when calling `datastore_service_create_datastore`")  # noqa: E501
        # verify the required parameter 'body' is set
        if self.api_client.client_side_validation and ('body' not in params or
                                                       params['body'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `body` when calling `datastore_service_create_datastore`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'spec_cluster_id' in params:
            path_params['spec.clusterId'] = params['spec_cluster_id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{spec.clusterId}/datastores', 'POST',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1CreateDatastoreResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def datastore_service_create_datastore_presigned_urls(self, cluster_id: 'str', datastore_id: 'str', body: 'Body2', **kwargs) -> 'V1CreateDatastorePresignedUrlsResponse':  # noqa: E501
        """CreateDatastorePresignedUrls returns a set of presigned urls which can be used to upload files directly to the platforms backing storage server.  # noqa: E501

        After all parts of an object have been uploaded from the client, the client must make a call to the CompleteDatastorePresignedUrlsUpload endpoint in order to mark the files as available in the datastore. Failure to finalize an object will silently cause that object to go \"missing\" when the datastore is attached to a compute service.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_create_datastore_presigned_urls(cluster_id, datastore_id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: the cluster the datastore lives on. (required)
        :param str datastore_id: the id of the datastore we are uploading to. (required)
        :param Body2 body: (required)
        :return: V1CreateDatastorePresignedUrlsResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.datastore_service_create_datastore_presigned_urls_with_http_info(cluster_id, datastore_id, body, **kwargs)  # noqa: E501
        else:
            (data) = self.datastore_service_create_datastore_presigned_urls_with_http_info(cluster_id, datastore_id, body, **kwargs)  # noqa: E501
            return data

    def datastore_service_create_datastore_presigned_urls_with_http_info(self, cluster_id: 'str', datastore_id: 'str', body: 'Body2', **kwargs) -> 'V1CreateDatastorePresignedUrlsResponse':  # noqa: E501
        """CreateDatastorePresignedUrls returns a set of presigned urls which can be used to upload files directly to the platforms backing storage server.  # noqa: E501

        After all parts of an object have been uploaded from the client, the client must make a call to the CompleteDatastorePresignedUrlsUpload endpoint in order to mark the files as available in the datastore. Failure to finalize an object will silently cause that object to go \"missing\" when the datastore is attached to a compute service.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_create_datastore_presigned_urls_with_http_info(cluster_id, datastore_id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: the cluster the datastore lives on. (required)
        :param str datastore_id: the id of the datastore we are uploading to. (required)
        :param Body2 body: (required)
        :return: V1CreateDatastorePresignedUrlsResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['cluster_id', 'datastore_id', 'body']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method datastore_service_create_datastore_presigned_urls" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'cluster_id' is set
        if self.api_client.client_side_validation and ('cluster_id' not in params or
                                                       params['cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `cluster_id` when calling `datastore_service_create_datastore_presigned_urls`")  # noqa: E501
        # verify the required parameter 'datastore_id' is set
        if self.api_client.client_side_validation and ('datastore_id' not in params or
                                                       params['datastore_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `datastore_id` when calling `datastore_service_create_datastore_presigned_urls`")  # noqa: E501
        # verify the required parameter 'body' is set
        if self.api_client.client_side_validation and ('body' not in params or
                                                       params['body'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `body` when calling `datastore_service_create_datastore_presigned_urls`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'cluster_id' in params:
            path_params['clusterId'] = params['cluster_id']  # noqa: E501
        if 'datastore_id' in params:
            path_params['datastoreId'] = params['datastore_id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{clusterId}/datastores/{datastoreId}/presigned-urls', 'POST',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1CreateDatastorePresignedUrlsResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def datastore_service_delete_datastore(self, cluster_id: 'str', id: 'str', **kwargs) -> 'V1DeleteDatastoreResponse':  # noqa: E501
        """datastore_service_delete_datastore  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_delete_datastore(cluster_id, id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str id: (required)
        :return: V1DeleteDatastoreResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.datastore_service_delete_datastore_with_http_info(cluster_id, id, **kwargs)  # noqa: E501
        else:
            (data) = self.datastore_service_delete_datastore_with_http_info(cluster_id, id, **kwargs)  # noqa: E501
            return data

    def datastore_service_delete_datastore_with_http_info(self, cluster_id: 'str', id: 'str', **kwargs) -> 'V1DeleteDatastoreResponse':  # noqa: E501
        """datastore_service_delete_datastore  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_delete_datastore_with_http_info(cluster_id, id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str id: (required)
        :return: V1DeleteDatastoreResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['cluster_id', 'id']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method datastore_service_delete_datastore" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'cluster_id' is set
        if self.api_client.client_side_validation and ('cluster_id' not in params or
                                                       params['cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `cluster_id` when calling `datastore_service_delete_datastore`")  # noqa: E501
        # verify the required parameter 'id' is set
        if self.api_client.client_side_validation and ('id' not in params or
                                                       params['id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `id` when calling `datastore_service_delete_datastore`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'cluster_id' in params:
            path_params['clusterId'] = params['cluster_id']  # noqa: E501
        if 'id' in params:
            path_params['id'] = params['id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{clusterId}/datastores/{id}', 'DELETE',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1DeleteDatastoreResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def datastore_service_get_datastore(self, cluster_id: 'str', id: 'str', **kwargs) -> 'V1GetDatastoreResponse':  # noqa: E501
        """datastore_service_get_datastore  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_get_datastore(cluster_id, id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str id: (required)
        :return: V1GetDatastoreResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.datastore_service_get_datastore_with_http_info(cluster_id, id, **kwargs)  # noqa: E501
        else:
            (data) = self.datastore_service_get_datastore_with_http_info(cluster_id, id, **kwargs)  # noqa: E501
            return data

    def datastore_service_get_datastore_with_http_info(self, cluster_id: 'str', id: 'str', **kwargs) -> 'V1GetDatastoreResponse':  # noqa: E501
        """datastore_service_get_datastore  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_get_datastore_with_http_info(cluster_id, id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str id: (required)
        :return: V1GetDatastoreResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['cluster_id', 'id']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method datastore_service_get_datastore" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'cluster_id' is set
        if self.api_client.client_side_validation and ('cluster_id' not in params or
                                                       params['cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `cluster_id` when calling `datastore_service_get_datastore`")  # noqa: E501
        # verify the required parameter 'id' is set
        if self.api_client.client_side_validation and ('id' not in params or
                                                       params['id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `id` when calling `datastore_service_get_datastore`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'cluster_id' in params:
            path_params['clusterId'] = params['cluster_id']  # noqa: E501
        if 'id' in params:
            path_params['id'] = params['id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{clusterId}/datastores/{id}', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1GetDatastoreResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def datastore_service_list_datastore_events(self, cluster_id: 'str', id: 'str', **kwargs) -> 'V1ListDatastoreEventsResponse':  # noqa: E501
        """datastore_service_list_datastore_events  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_list_datastore_events(cluster_id, id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str id: (required)
        :return: V1ListDatastoreEventsResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.datastore_service_list_datastore_events_with_http_info(cluster_id, id, **kwargs)  # noqa: E501
        else:
            (data) = self.datastore_service_list_datastore_events_with_http_info(cluster_id, id, **kwargs)  # noqa: E501
            return data

    def datastore_service_list_datastore_events_with_http_info(self, cluster_id: 'str', id: 'str', **kwargs) -> 'V1ListDatastoreEventsResponse':  # noqa: E501
        """datastore_service_list_datastore_events  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_list_datastore_events_with_http_info(cluster_id, id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str id: (required)
        :return: V1ListDatastoreEventsResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['cluster_id', 'id']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method datastore_service_list_datastore_events" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'cluster_id' is set
        if self.api_client.client_side_validation and ('cluster_id' not in params or
                                                       params['cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `cluster_id` when calling `datastore_service_list_datastore_events`")  # noqa: E501
        # verify the required parameter 'id' is set
        if self.api_client.client_side_validation and ('id' not in params or
                                                       params['id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `id` when calling `datastore_service_list_datastore_events`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'cluster_id' in params:
            path_params['clusterId'] = params['cluster_id']  # noqa: E501
        if 'id' in params:
            path_params['id'] = params['id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{clusterId}/datastores/{id}/events', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1ListDatastoreEventsResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def datastore_service_list_datastore_files(self, id: 'str', **kwargs) -> 'V1ListDatastoreFilesResponse':  # noqa: E501
        """ListDatastoreFiles returns a paginated list of files in the datastore  # noqa: E501

        This API is cluster agnostic in order to allow for multi-cluster datastores in the future.  N.B. This API is currently unavailable for datastores created from datastore source types which are OBJECT_STORE_REMOTE_REFERENCE (ie. 'no-copy') datastores due to limitations around authentication to external s3 accounts in the controlplane. This will be lifted in the future once we addresss these concerns.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_list_datastore_files(id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str id: (required)
        :param str page_token:
        :param str page_size: limit for returned file info. Default (if not set) = 100.
        :return: V1ListDatastoreFilesResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.datastore_service_list_datastore_files_with_http_info(id, **kwargs)  # noqa: E501
        else:
            (data) = self.datastore_service_list_datastore_files_with_http_info(id, **kwargs)  # noqa: E501
            return data

    def datastore_service_list_datastore_files_with_http_info(self, id: 'str', **kwargs) -> 'V1ListDatastoreFilesResponse':  # noqa: E501
        """ListDatastoreFiles returns a paginated list of files in the datastore  # noqa: E501

        This API is cluster agnostic in order to allow for multi-cluster datastores in the future.  N.B. This API is currently unavailable for datastores created from datastore source types which are OBJECT_STORE_REMOTE_REFERENCE (ie. 'no-copy') datastores due to limitations around authentication to external s3 accounts in the controlplane. This will be lifted in the future once we addresss these concerns.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_list_datastore_files_with_http_info(id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str id: (required)
        :param str page_token:
        :param str page_size: limit for returned file info. Default (if not set) = 100.
        :return: V1ListDatastoreFilesResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['id', 'page_token', 'page_size']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method datastore_service_list_datastore_files" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'id' is set
        if self.api_client.client_side_validation and ('id' not in params or
                                                       params['id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `id` when calling `datastore_service_list_datastore_files`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'id' in params:
            path_params['id'] = params['id']  # noqa: E501

        query_params = []
        if 'page_token' in params:
            query_params.append(('pageToken', params['page_token']))  # noqa: E501
        if 'page_size' in params:
            query_params.append(('pageSize', params['page_size']))  # noqa: E501

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/datastores/{id}/files', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1ListDatastoreFilesResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def datastore_service_list_datastores(self, cluster_id: 'str', **kwargs) -> 'V1ListDatastoresResponse':  # noqa: E501
        """datastore_service_list_datastores  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_list_datastores(cluster_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str page_token:
        :param str limit:
        :param list[str] user_ids:
        :param bool available:
        :return: V1ListDatastoresResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.datastore_service_list_datastores_with_http_info(cluster_id, **kwargs)  # noqa: E501
        else:
            (data) = self.datastore_service_list_datastores_with_http_info(cluster_id, **kwargs)  # noqa: E501
            return data

    def datastore_service_list_datastores_with_http_info(self, cluster_id: 'str', **kwargs) -> 'V1ListDatastoresResponse':  # noqa: E501
        """datastore_service_list_datastores  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_list_datastores_with_http_info(cluster_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str page_token:
        :param str limit:
        :param list[str] user_ids:
        :param bool available:
        :return: V1ListDatastoresResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['cluster_id', 'page_token', 'limit', 'user_ids', 'available']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method datastore_service_list_datastores" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'cluster_id' is set
        if self.api_client.client_side_validation and ('cluster_id' not in params or
                                                       params['cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `cluster_id` when calling `datastore_service_list_datastores`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'cluster_id' in params:
            path_params['clusterId'] = params['cluster_id']  # noqa: E501

        query_params = []
        if 'page_token' in params:
            query_params.append(('pageToken', params['page_token']))  # noqa: E501
        if 'limit' in params:
            query_params.append(('limit', params['limit']))  # noqa: E501
        if 'user_ids' in params:
            query_params.append(('userIds', params['user_ids']))  # noqa: E501
            collection_formats['userIds'] = 'multi'  # noqa: E501
        if 'available' in params:
            query_params.append(('available', params['available']))  # noqa: E501

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{clusterId}/datastores', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1ListDatastoresResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def datastore_service_mark_datastore_upload_complete(self, cluster_id: 'str', datastore_id: 'str', **kwargs) -> 'V1MarkDatastoreUploadCompleteResponse':  # noqa: E501
        """MarkDatastoreUploadComplete informs grid that all data files have been uploaded.  # noqa: E501

        For the sake of consistency, this method **MUST** be called regardless of the data source type used when creating the datastore, even if the datastore is being created from a remote source such as an S3 bucket or HTTP URL download. **N.B.** Posting to this endpoint disables the ability to generate new presigned URLs for uploading files ot this datastore.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_mark_datastore_upload_complete(cluster_id, datastore_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: the cluster the datastore lives on. (required)
        :param str datastore_id: the id of the datastore. (required)
        :return: V1MarkDatastoreUploadCompleteResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.datastore_service_mark_datastore_upload_complete_with_http_info(cluster_id, datastore_id, **kwargs)  # noqa: E501
        else:
            (data) = self.datastore_service_mark_datastore_upload_complete_with_http_info(cluster_id, datastore_id, **kwargs)  # noqa: E501
            return data

    def datastore_service_mark_datastore_upload_complete_with_http_info(self, cluster_id: 'str', datastore_id: 'str', **kwargs) -> 'V1MarkDatastoreUploadCompleteResponse':  # noqa: E501
        """MarkDatastoreUploadComplete informs grid that all data files have been uploaded.  # noqa: E501

        For the sake of consistency, this method **MUST** be called regardless of the data source type used when creating the datastore, even if the datastore is being created from a remote source such as an S3 bucket or HTTP URL download. **N.B.** Posting to this endpoint disables the ability to generate new presigned URLs for uploading files ot this datastore.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_mark_datastore_upload_complete_with_http_info(cluster_id, datastore_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: the cluster the datastore lives on. (required)
        :param str datastore_id: the id of the datastore. (required)
        :return: V1MarkDatastoreUploadCompleteResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['cluster_id', 'datastore_id']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method datastore_service_mark_datastore_upload_complete" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'cluster_id' is set
        if self.api_client.client_side_validation and ('cluster_id' not in params or
                                                       params['cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `cluster_id` when calling `datastore_service_mark_datastore_upload_complete`")  # noqa: E501
        # verify the required parameter 'datastore_id' is set
        if self.api_client.client_side_validation and ('datastore_id' not in params or
                                                       params['datastore_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `datastore_id` when calling `datastore_service_mark_datastore_upload_complete`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'cluster_id' in params:
            path_params['clusterId'] = params['cluster_id']  # noqa: E501
        if 'datastore_id' in params:
            path_params['datastoreId'] = params['datastore_id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{clusterId}/datastores/{datastoreId}/completed', 'POST',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1MarkDatastoreUploadCompleteResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def datastore_service_update_datastore(self, spec_cluster_id: 'str', id: 'str', body: 'Body6', **kwargs) -> 'V1UpdateDatastoreResponse':  # noqa: E501
        """datastore_service_update_datastore  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_update_datastore(spec_cluster_id, id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str spec_cluster_id: (required)
        :param str id: (required)
        :param Body6 body: (required)
        :return: V1UpdateDatastoreResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.datastore_service_update_datastore_with_http_info(spec_cluster_id, id, body, **kwargs)  # noqa: E501
        else:
            (data) = self.datastore_service_update_datastore_with_http_info(spec_cluster_id, id, body, **kwargs)  # noqa: E501
            return data

    def datastore_service_update_datastore_with_http_info(self, spec_cluster_id: 'str', id: 'str', body: 'Body6', **kwargs) -> 'V1UpdateDatastoreResponse':  # noqa: E501
        """datastore_service_update_datastore  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_update_datastore_with_http_info(spec_cluster_id, id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str spec_cluster_id: (required)
        :param str id: (required)
        :param Body6 body: (required)
        :return: V1UpdateDatastoreResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['spec_cluster_id', 'id', 'body']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method datastore_service_update_datastore" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'spec_cluster_id' is set
        if self.api_client.client_side_validation and ('spec_cluster_id' not in params or
                                                       params['spec_cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `spec_cluster_id` when calling `datastore_service_update_datastore`")  # noqa: E501
        # verify the required parameter 'id' is set
        if self.api_client.client_side_validation and ('id' not in params or
                                                       params['id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `id` when calling `datastore_service_update_datastore`")  # noqa: E501
        # verify the required parameter 'body' is set
        if self.api_client.client_side_validation and ('body' not in params or
                                                       params['body'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `body` when calling `datastore_service_update_datastore`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'spec_cluster_id' in params:
            path_params['spec.clusterId'] = params['spec_cluster_id']  # noqa: E501
        if 'id' in params:
            path_params['id'] = params['id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{spec.clusterId}/datastores/{id}', 'PUT',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1UpdateDatastoreResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING

import six

from grid.openapi.configuration import Configuration

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *

class V1InstanceType(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'availability': 'InstanceTypeAvailability',
        'cpu': 'str',
        'gpu': 'str',
        'hourly_cost': 'float',
        'memory': 'str',
        'name': 'str',
        'overprovisioned_ondemand_count': 'int',
        'spot': 'bool'
    }

    attribute_map = {
        'availability': 'availability',
        'cpu': 'cpu',
        'gpu': 'gpu',
        'hourly_cost': 'hourlyCost',
        'memory': 'memory',
        'name': 'name',
        'overprovisioned_ondemand_count': 'overprovisionedOndemandCount',
        'spot': 'spot'
    }

    def __init__(self, availability: 'InstanceTypeAvailability' = None, cpu: 'str' = None, gpu: 'str' = None, hourly_cost: 'float' = None, memory: 'str' = None, name: 'str' = None, overprovisioned_ondemand_count: 'int' = None, spot: 'bool' = None, _configuration=None):  # noqa: E501
        """V1InstanceType - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._availability = None
        self._cpu = None
        self._gpu = None
        self._hourly_cost = None
        self._memory = None
        self._name = None
        self._overprovisioned_ondemand_count = None
        self._spot = None
        self.discriminator = None

        if availability is not None:
            self.availability = availability
        if cpu is not None:
            self.cpu = cpu
        if gpu is not None:
            self.gpu = gpu
        if hourly_cost is not None:
            self.hourly_cost = hourly_cost
        if memory is not None:
            self.memory = memory
        if name is not None:
            self.name = name
        if overprovisioned_ondemand_count is not None:
            self.overprovisioned_ondemand_count = overprovisioned_ondemand_count
        if spot is not None:
            self.spot = spot

    @property
    def availability(self) -> 'InstanceTypeAvailability':
        """Gets the availability of this V1InstanceType.  # noqa: E501


        :return: The availability of this V1InstanceType.  # noqa: E501
        :rtype: InstanceTypeAvailability
        """
        return self._availability

    @availability.setter
    def availability(self, availability: 'InstanceTypeAvailability'):
        """Sets the availability of this V1InstanceType.


        :param availability: The availability of this V1InstanceType.  # noqa: E501
        :type: InstanceTypeAvailability
        """

        self._availability = availability

    @property
    def cpu(self) -> 'str':
        """Gets the cpu of this V1InstanceType.  # noqa: E501


        :return: The cpu of this V1InstanceType.  # noqa: E501
        :rtype: str
        """
        return self._cpu

    @cpu.setter
    def cpu(self, cpu: 'str'):
        """Sets the cpu of this V1InstanceType.


        :param cpu: The cpu of this V1InstanceType.  # noqa: E501
        :type: str
        """

        self._cpu = cpu

    @property
    def gpu(self) -> 'str':
        """Gets the gpu of this V1InstanceType.  # noqa: E501


        :return: The gpu of this V1InstanceType.  # noqa: E501
        :rtype: str
        """
        return self._gpu

    @gpu.setter
    def gpu(self, gpu: 'str'):
        """Sets the gpu of this V1InstanceType.


        :param gpu: The gpu of this V1InstanceType.  # noqa: E501
        :type: str
        """

        self._gpu = gpu

    @property
    def hourly_cost(self) -> 'float':
        """Gets the hourly_cost of this V1InstanceType.  # noqa: E501


        :return: The hourly_cost of this V1InstanceType.  # noqa: E501
        :rtype: float
        """
        return self._hourly_cost

    @hourly_cost.setter
    def hourly_cost(self, hourly_cost: 'float'):
        """Sets the hourly_cost of this V1InstanceType.


        :param hourly_cost: The hourly_cost of this V1InstanceType.  # noqa: E501
        :type: float
        """

        self._hourly_cost = hourly_cost

    @property
    def memory(self) -> 'str':
        """Gets the memory of this V1InstanceType.  # noqa: E501


        :return: The memory of this V1InstanceType.  # noqa: E501
        :rtype: str
        """
        return self._memory

    @memory.setter
    def memory(self, memory: 'str'):
        """Sets the memory of this V1InstanceType.


        :param memory: The memory of this V1InstanceType.  # noqa: E501
        :type: str
        """

        self._memory = memory

    @property
    def name(self) -> 'str':
        """Gets the name of this V1InstanceType.  # noqa: E501


        :return: The name of this V1InstanceType.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name: 'str'):
        """Sets the name of this V1InstanceType.


        :param name: The name of this V1InstanceType.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def overprovisioned_ondemand_count(self) -> 'int':
        """Gets the overprovisioned_ondemand_count of this V1InstanceType.  # noqa: E501


        :return: The overprovisioned_ondemand_count of this V1InstanceType.  # noqa: E501
        :rtype: int
        """
        return self._overprovisioned_ondemand_count

    @overprovisioned_ondemand_count.setter
    def overprovisioned_ondemand_count(self, overprovisioned_ondemand_count: 'int'):
        """Sets the overprovisioned_ondemand_count of this V1InstanceType.


        :param overprovisioned_ondemand_count: The overprovisioned_ondemand_count of this V1InstanceType.  # noqa: E501
        :type: int
        """

        self._overprovisioned_ondemand_count = overprovisioned_ondemand_count

    @property
    def spot(self) -> 'bool':
        """Gets the spot of this V1InstanceType.  # noqa: E501


        :return: The spot of this V1InstanceType.  # noqa: E501
        :rtype: bool
        """
        return self._spot

    @spot.setter
    def spot(self, spot: 'bool'):
        """Sets the spot of this V1InstanceType.


        :param spot: The spot of this V1InstanceType.  # noqa: E501
        :type: bool
        """

        self._spot = spot

    def to_dict(self) -> dict:
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(V1InstanceType, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other: 'V1InstanceType') -> bool:
        """Returns true if both objects are equal"""
        if not isinstance(other, V1InstanceType):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other: 'V1InstanceType') -> bool:
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1InstanceType):
            return True

        return self.to_dict() != other.to_dict()



from dataclasses import dataclass, field
from datetime import date
from typing import Optional

import dateutil.parser
from dataclasses_json import dataclass_json, LetterCase, config
from dataclasses_json.mm import _IsoField

from grid.sdk import env
from grid.sdk.client import create_swagger_client
from grid.sdk.rest import GridRestClient
from grid.sdk.rest.exceptions import throw_with_message
from grid.sdk.utils.arnparser import arnparse, MalformedArnError


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class AWSBucketAccessSpec:
    role_arn: str
    external_id: str

    def validate(self):
        """Raises ValueError if the values are not valid for the bucket access spec.
        """
        if self.role_arn == "":
            raise ValueError("role_arn cannot be empty string")

        if self.external_id == "":
            raise ValueError("external_id cannot be empty string")

        try:
            arn = arnparse(self.role_arn)
        except MalformedArnError as e:
            raise ValueError(
                f"the role_arn: {self.role_arn} is not a valid AWS ARN. Please refer to "
                f"https://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html "
                f"for what a valid value looks like"
            ) from e

        if not arn.resource.startswith("grid-s3-access-"):
            raise ValueError(
                f"role_arn `{self.role_arn}` resource-id value `{arn.resource}` role "
                f"name does not start with the required prefix `grid-s3-access-`"
            )

        if arn.resource == "grid-s3-access-":
            raise ValueError(
                f"role_arn `{self.role_arn}` resource-id value `{arn.resource}` suffix must "
                f"consist of atleast one character after the required prefix `grid-s3-access-`"
            )


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class GCPBucketAccessSpec:
    pass


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class AzureBucketAccessSpec:
    pass


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class BucketAccessCredential:
    # one_of field
    aws: Optional[AWSBucketAccessSpec] = field(
        default=None,
        metadata=config(exclude=lambda x: x is None),
    )
    # one_of field
    gcp: Optional[GCPBucketAccessSpec] = field(
        default=None,
        metadata=config(exclude=lambda x: x is None),
    )
    # one_of field
    azure: Optional[AzureBucketAccessSpec] = field(
        default=None,
        metadata=config(exclude=lambda x: x is None),
    )

    # protobuf json serialization encodes/decodes to RFC 3339 format
    # see https://pkg.go.dev/google.golang.org/protobuf/types/known/timestamppb#Timestamp
    creation_timestamp: Optional[date] = field(
        default=None,
        metadata=config(
            encoder=lambda x: x.strftime('%Y-%m-%dT%H:%M:%S.%fZ') if x is not None else None,
            decoder=lambda x: dateutil.parser.parse(x) if x is not None else None,
            mm_field=_IsoField(),
            exclude=lambda x: x is None,
        )
    )
    # protobuf json serialization encodes/decodes to RFC 3339 format
    # see https://pkg.go.dev/google.golang.org/protobuf/types/known/timestamppb#Timestamp
    update_timestamp: Optional[date] = field(
        default=None,
        metadata=config(
            encoder=lambda x: x.strftime('%Y-%m-%dT%H:%M:%S.%fZ') if x is not None else None,
            decoder=lambda x: dateutil.parser.parse(x) if x is not None else None,
            mm_field=_IsoField(),
            exclude=lambda x: x is None,
        )
    )
    # protobuf json serialization encodes/decodes to RFC 3339 format
    # see https://pkg.go.dev/google.golang.org/protobuf/types/known/timestamppb#Timestamp
    deletion_timestamp: Optional[date] = field(
        default=None,
        metadata=config(
            encoder=lambda x: x.strftime('%Y-%m-%dT%H:%M:%S.%fZ') if x is not None else None,
            decoder=lambda x: dateutil.parser.parse(x) if x is not None else None,
            mm_field=_IsoField(),
            exclude=lambda x: x is None,
        )
    )

    phase: Optional[str] = field(
        default=None,
        metadata=config(exclude=lambda x: x is None),
    )
    reason: Optional[str] = field(
        default=None,
        metadata=config(exclude=lambda x: x is None),
    )
    message: Optional[str] = field(
        default=None,
        metadata=config(exclude=lambda x: x is None),
    )
    retry_count: Optional[int] = field(
        default=None,
        metadata=config(exclude=lambda x: x is None),
    )

    def name(self):
        if self.aws is not None:
            return arnparse(self.aws.role_arn).resource
        elif self.gcp is not None:
            raise NotImplementedError("gcp credentials are not currently supported")
        elif self.azure is not None:
            raise NotImplementedError("gcp credentials are not currently supported")
        else:
            raise RuntimeError("internal error occured - no provider access credential specified")


@throw_with_message
def get_cluster_account_id(cluster_id: str) -> str:
    c = GridRestClient(create_swagger_client())
    resp = c.cluster_service_get_cluster_cloud_provider_account_id(id=cluster_id)
    return resp.account_id


def get_cluster_controlplane_aws_account_id(cluster_id: str) -> str:
    """get the aws account principle which the controlplane assumes when making requests

    TODO
    ----
    This is a dummy method right now. Actually implement a way to get this from the backend!

    Since there's no easy way to get the account id of the grid-role principle account
    the controlplane assumes, we're going to hardcode this to the byoc-sandbox and grid-ai
    AWS accounts for staging/testing/prod. This isn't a problem until we have on-prem
    controlplane deployments as all BYOC clusters use the same aws principle account (i.e. ours!)
    """
    if env.TESTING or cluster_id == 'staging-3':
        return "158793097533"  # byoc-sandbox

    return "302180240179"  # grid-ai



import subprocess
from pathlib import Path
from typing import Optional

import click

from grid.sdk import Session, env
from grid.cli import rich_click
from grid.cli.cli.grid_run import _check_run_name_is_valid, _get_instance_types, _resolve_instance_type_nickname
from grid.cli.cli.utilities import validate_disk_size_callback
from grid.cli.client import Grid
from grid.cli.types import ObservableType
from grid.sdk.datastores import fetch_datastore
from grid.sdk.utils.name_generator import unique_name
import yaspin

WARNING_STR = click.style('WARNING', fg='yellow')
SUCCESS_MARK = click.style("✔", fg='green')
FAIL_MARK = click.style('✘', fg='red')


@rich_click.group(invoke_without_command=True)
@click.pass_context
@click.option(
    '--global',
    'is_global',
    type=bool,
    is_flag=True,
    help='Fetch sessions from everyone in the team when flag is passed'
)
def session(ctx, is_global: bool) -> None:
    """
    Contains a grouping of commands to manage sessions workflows.

    Executing the `grid session` command without any further arguments
    or commands renders a list of all sessions registered to your Grid
    user account.
    """
    client = Grid()
    if ctx.invoked_subcommand is None:
        # Get the status of the interactive observables.
        click.echo(f"Loading sessions in {env.CONTEXT}")
        kind = ObservableType.INTERACTIVE
        client.status(kind=kind, follow=False, is_global=is_global)
    elif is_global:
        click.echo(f"{WARNING_STR}: --global flag doesn't have any effect when invoked with a subcommand")


@session.command()
@click.option('--cluster', 'cluster', type=str, required=False, default=env.CONTEXT, help='Cluster to run on')
@click.option(
    '--instance_type',
    'instance_type',
    type=str,
    default='m5a.large',
    callback=_resolve_instance_type_nickname,
    help='Instance type to start session in.',
)
@click.option(
    '--use_spot',
    'use_spot',
    is_flag=True,
    required=False,
    default=False,
    help='Use spot instance. The spot instances, or preemptive instance can be shut down at will'
)
@click.option(
    '--disk_size',
    'disk_size',
    type=int,
    required=False,
    default=200,
    callback=validate_disk_size_callback,
    help='The disk size in GB to allocate to the session.'
)
@click.option(
    '--datastore_name',
    'datastore_name',
    type=str,
    required=False,
    default=None,
    help='Datastore name to be mounted in the session.'
)
@click.option(
    '--datastore_version',
    'datastore_version',
    type=int,
    required=False,
    default=None,
    help='Datastore version to be mounted in the session.'
)
@click.option(
    '--datastore_mount_dir',
    'datastore_mount_dir',
    type=str,
    required=False,
    default=None,
    help='Absolute path to mount Datastore in the session (defaults to /datastores/<datastore-name>).'
)
@click.option('--config', 'config', type=Path, required=False, default=None, help='Path to Grid config YML')
@click.option(
    '--name', 'name', type=str, required=False, help='Name for this session', callback=_check_run_name_is_valid
)
def create(
    name: Optional[str], cluster: Optional[str], instance_type: str, datastore_name: Optional[str],
    datastore_version: Optional[int], datastore_mount_dir: Optional[str], disk_size: int, use_spot: bool,
    config: Optional[Path]
) -> None:
    """Creates a new interactive session with NAME.

    Interactive sessions are optimized for development activites (before executing
    hyperparemeter sweeps in a Run). Once created, sessions can be accessed via
    VSCode, Jupyter-lab, or SSH interfaces.

    Grid manages the installation of any/all core libraries, drivers, and interfaces
    to the outside world. Sessions can be run on anything from a small 2 CPU core +
    4GB memory instance to a monster machine with 96 CPU cores + 824 GB memory + eight
    V100 GPUs + 40 GBPS network bandwidth (no, those values aren't typos!); or really
    anything in between.
    """
    # make a fun random name when user does not pass in a name
    if name is None:
        name = unique_name()
    spinner = yaspin.yaspin(text=f"Submitting Session {name} ...", color="yellow")
    try:
        spinner.start()
        # process datastore specification args
        dstore = None
        if datastore_name:
            dstore = fetch_datastore(datastore_name, datastore_version, cluster)

        sess = Session(
            name=name,
            instance_type=instance_type,
            disk_size_gb=disk_size,
            use_spot=use_spot,
            cluster_id=cluster,
            datastore=dstore,
            datastore_mount_dir=datastore_mount_dir,
            config_file=config,
        )
        if sess.exists is True:
            raise click.ClickException(f"A session with the name: {name} already exists. Please specify a unique name.")

        click.echo("Creating Interactive session ...", color="yellow")
        sess.start()
        click.echo(
            f"""
        {SUCCESS_MARK} Interactive session created!

        `grid status` to list all runs and interactive sessions.
        `grid status {name}` to see the status for this interactive session.

        ----------------------
        Submission summary
        ----------------------
        name:                    {name}
        instance_type:           {sess.instance_type}
        cluster_id:              {sess.cluster_id}
        datastore_name:          {datastore_name}
        datastore_version:       {datastore_version}
        datastore_mount_dir:     {sess.datastore_mount_dir}
        use_spot:                {sess.use_spot}
        """
        )
        click.echo(f"Interactive session {name} is spinning up.")
    except Exception as e:
        click.echo(f'{FAIL_MARK} {e}')
        raise click.ClickException(str(e))
    finally:
        spinner.stop()


@session.command()
@rich_click.argument('session_name', type=str, nargs=1)
def pause(session_name: str) -> None:
    """Pauses a session identified by the SESSION_NAME.

    Pausing a session stops the running instance (and any computations being
    performed on it - be sure to save your work!) and and billing of your account
    for the machine. The session can be resumed at a later point with all your
    persisted files and saved work unchanged.
    """
    click.echo("Pausing Interactive session ...", color="yellow")

    try:
        sess = Session(name=session_name)
        if not sess.exists:
            raise RuntimeError(
                f"Session {session_name} does not exist in {env.CONTEXT}. "
                f"If you are pausing a session in another cluster, set "
                f"the default cluster first using "
                f"`grid user set-default-cluster <cluster_name>`."
            )
        sess.pause()
        click.echo(SUCCESS_MARK)
        click.echo(f'Interactive session {session_name} has been paused successfully.')
    except Exception as e:
        click.echo("✘", color="red")
        raise click.ClickException(f"Failed to pause interactive session: '{session_name}': {e}")


@session.command()
@rich_click.argument('session_name', type=str, nargs=1)
def resume(session_name: str) -> None:
    """Resumes a session identified by SESSION_NAME.
    """
    click.echo("Resuming Interactive session ...", color="yellow")

    try:
        sess = Session(name=session_name)
        if not sess.exists:
            raise RuntimeError(
                f"Session {session_name} does not exist in {env.CONTEXT}. "
                f"If you are pausing a session in another cluster, set "
                f"the default cluster first using "
                f"`grid user set-default-cluster <cluster_name>`."
            )
        sess.start()
        click.echo(f'{SUCCESS_MARK} Interactive session: {session_name} successfully began resuming.')
        click.echo('Note: Depending on instance-type selected, it may take up to 10 minutes to become available.')
    except Exception as e:
        click.echo(f"{FAIL_MARK} {e}")
        raise click.ClickException(f"Failed to resume interactive session: '{session_name}'")


@session.command()
@rich_click.argument('session_name', type=str, nargs=1)
def delete(session_name: str) -> None:
    """Deletes a session identified by SESSION_NAME.

    Deleting a session will stop the running instance (and any computations being
    performed on it) and billing of your account. All work done on the machine
    is permenantly removed, including all/any saved files, code, or downloaded
    data (assuming the source of the data was not a grid datastore - datastore
    data is not deleted).
    """
    click.echo("Deleting Interactive node ...", color="yellow")

    try:
        sess = Session(name=session_name)
        if not sess.exists:
            raise RuntimeError(
                f"Session {session_name} does not exist in {env.CONTEXT}. "
                f"If you are pausing a session in another cluster, set "
                f"the default cluster first using "
                f"`grid user set-default-cluster <cluster_name>`."
            )
        sess.delete()
        click.echo(f'{SUCCESS_MARK} Interactive node {session_name} has been deleted successfully.')
    except Exception as e:
        click.echo(f'{FAIL_MARK} {e}')
        raise click.ClickException(f"Failed to delete interactive session '{session_name}'")


@session.command()
@rich_click.argument('session_name', type=str, nargs=1, help='Name of the session to change')
@rich_click.argument(
    'instance_type', type=str, nargs=1, help='Instance type to change to', callback=_resolve_instance_type_nickname
)
@click.option(
    '--spot',
    type=bool,
    is_flag=True,
    default=None,
    show_default=True,
    help='Use a spot instance to launch the session'
)
@click.option(
    '--on_demand',
    '--on-demand',
    'on_demand',
    type=bool,
    is_flag=True,
    default=None,
    show_default=True,
    help='Use an on-demand instance to launch the session'
)
def change_instance_type(
    session_name: str, instance_type: str, spot: Optional[bool], on_demand: Optional[bool]
) -> None:
    """
    Change the instance type of a session; this allows you to upgrade
    or downgrade the compute capability of the session nodes while keeping
    all of your work in progress untouched.

    The session must be PAUSED in order for this command to succeed

    Specifying --spot allows you to change the instance to an interuptable
    spot instances (which come at a steap discount, but which can be
    interrupted and shut down at any point in time depending on cloud
    provider instance type demand).

    specifying --on_demand changes the instance to an on-demand type,
    which cannot be inturrupted but is more expensive.
    """
    click.echo("Changing Session Instance Type ...", color="yellow")

    if (spot is True) and (on_demand is True):
        raise click.ClickException('cannot pass both --spot and --on_demand flags to this command.')

    use_spot = True if spot else False if on_demand else None

    try:

        sess = Session(name=session_name)
        if not sess.exists:
            raise RuntimeError(
                f"Session {session_name} does not exist in {env.CONTEXT}. "
                f"If you are pausing a session in another cluster, set "
                f"the default cluster first using "
                f"`grid user set-default-cluster <cluster_name>`."
            )
        sess.change_instance_type(instance_type=instance_type, use_spot=use_spot)
        click.echo(f'{SUCCESS_MARK} Interactive session {session_name} instance type changed successfully.')
    except Exception as e:
        click.echo(f'{FAIL_MARK} {e}')
        raise click.ClickException(f"Failed to change session instance type '{session_name}'")


def _update_ssh_config_and_check_ixsession_status(ctx, param, value: str) -> int:
    """
    This updates the SSH config for interactive nodes and
    also checks if those nodes can be interacted with SSH
    before attempting an operation. This prevents SSHing into
    nodes that are not un a running state, e.g. paused or pending.

    This manages a section within user's ssh config file
    for all interactive nodes shh config details

    Afterwards you can use systems ssh & related utilities
    (sshfs, rsync, ansible, whatever) with interactive nodes
    directly

    The default file is ~/.ssh/config and  can be changed via
    envvar GRID_SSH_CONFIG

    Returns
    --------
    value: str
        Unmodified value if valid
    """
    client = Grid()
    nodes = client.sync_ssh_config()

    click.echo(f"Sync config for {len(nodes)} interactive nodes")

    target_ixsession = None
    for node in nodes:
        if node["name"] == value:
            target_ixsession = node

    # Check if interactive session exists at all
    if not target_ixsession:
        session_names = [n["name"] for n in nodes]
        raise click.BadArgumentUsage(
            f"Interactive session {value} does not exist. "
            f"Available Interactive Sessions are: {', '.join(session_names)}"
        )

    # Check if the node is in 'running' status
    if target_ixsession["status"] != "running":
        running_ixsessions = [n["name"] for n in nodes if n["status"] == "running"]
        raise click.BadArgumentUsage(
            f"Interactive session {value} is not ready. "
            f"Sessions that are ready to SSH are: {', '.join(running_ixsessions)}"
        )

    return value


@session.command()
@rich_click.argument(
    'node_name',
    type=str,
    callback=_update_ssh_config_and_check_ixsession_status,
    help="The name of the node. This command executes ssh <node name>.",
)
@rich_click.argument('ssh_args', nargs=-1, type=click.UNPROCESSED, help="Arguments to be forwarded to the SSH command.")
def ssh(node_name, ssh_args):
    """SSH into the interactive node identified by NODE_NAME.

    If you'd like the full power of ssh, you can use any ssh client and
    do `ssh <node_name>`. This command is stripped down version of it.

    Example:

        1. Path to custom key:

        grid session ssh satisfied-rabbit-962 -- -i ~/.ssh/my-key

        2. Custom ssh option:

        grid session ssh satisfied-rabbit-962 -- -o "StrictHostKeyChecking accept-new"
    """
    subprocess.check_call(['ssh', node_name, *ssh_args])


@session.command()
@rich_click.argument('interactive_node', type=str, nargs=1, callback=_update_ssh_config_and_check_ixsession_status)
@rich_click.argument('mount_dir', type=str, nargs=1)
def mount(interactive_node, mount_dir):
    r"""Mount session directory to local. The session is identified by
    SESSION and MOUNT_DIR is a path to a directory on the local machine.

    To mount a filesystem use:
    ixNode:[dir] mountpoint

    Examples:
        # Mounts the home directory on the interactive node in dir data
        grid session mount bluberry-122 ./data

        # mounts ~/data directory on the interactive node to ./data
        grid session mount bluberry-122:~/data ./data

    To unmount it:
      fusermount3 -u mountpoint   # Linux
      umount mountpoint           # OS X, FreeBSD

    Under the hood this is just passing data to sshfs after syncing grid's interactive,
    i.e. this command is dumbed down sshfs

    """
    if ':' not in interactive_node:
        interactive_node += ":/home/jovyan"

    client = Grid()
    client.sync_ssh_config()

    try:
        subprocess.check_call(['sshfs', interactive_node, mount_dir])
    except FileNotFoundError:
        raise click.ClickException('Unable to mount: sshfs was not found')
    except subprocess.CalledProcessError as e:
        raise click.ClickException(f'Unable to mount: sshfs failed with code {e.returncode}')



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


from __future__ import absolute_import

import re  # noqa: F401
from typing import TYPE_CHECKING, Any

# python 2 and python 3 compatibility library
import six

from grid.openapi.api_client import ApiClient

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *


class CredentialServiceApi(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    Ref: https://github.com/swagger-api/swagger-codegen
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

    def credential_service_create_credential(self, credential_cluster_id: 'str', body: 'V1CreateCredentialRequest', **kwargs) -> 'V1Credential':  # noqa: E501
        """we return the credential (which may or may not be set depending on the credential type) this is done so that for credentials like the \"bucket access config\", data the user needs to configure (such as \"External ID\") can be used and the credential can be finalized by calling the UpdateCredential rpc with the RoleARN filled in.  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.credential_service_create_credential(credential_cluster_id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str credential_cluster_id: (required)
        :param V1CreateCredentialRequest body: (required)
        :return: V1Credential
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.credential_service_create_credential_with_http_info(credential_cluster_id, body, **kwargs)  # noqa: E501
        else:
            (data) = self.credential_service_create_credential_with_http_info(credential_cluster_id, body, **kwargs)  # noqa: E501
            return data

    def credential_service_create_credential_with_http_info(self, credential_cluster_id: 'str', body: 'V1CreateCredentialRequest', **kwargs) -> 'V1Credential':  # noqa: E501
        """we return the credential (which may or may not be set depending on the credential type) this is done so that for credentials like the \"bucket access config\", data the user needs to configure (such as \"External ID\") can be used and the credential can be finalized by calling the UpdateCredential rpc with the RoleARN filled in.  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.credential_service_create_credential_with_http_info(credential_cluster_id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str credential_cluster_id: (required)
        :param V1CreateCredentialRequest body: (required)
        :return: V1Credential
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['credential_cluster_id', 'body']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method credential_service_create_credential" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'credential_cluster_id' is set
        if self.api_client.client_side_validation and ('credential_cluster_id' not in params or
                                                       params['credential_cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `credential_cluster_id` when calling `credential_service_create_credential`")  # noqa: E501
        # verify the required parameter 'body' is set
        if self.api_client.client_side_validation and ('body' not in params or
                                                       params['body'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `body` when calling `credential_service_create_credential`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'credential_cluster_id' in params:
            path_params['credential.clusterId'] = params['credential_cluster_id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{credential.clusterId}/credentials', 'POST',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1Credential',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def credential_service_delete_credential(self, cluster_id: 'str', id: 'str', **kwargs) -> 'V1DeleteCredentialResponse':  # noqa: E501
        """credential_service_delete_credential  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.credential_service_delete_credential(cluster_id, id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str id: (required)
        :return: V1DeleteCredentialResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.credential_service_delete_credential_with_http_info(cluster_id, id, **kwargs)  # noqa: E501
        else:
            (data) = self.credential_service_delete_credential_with_http_info(cluster_id, id, **kwargs)  # noqa: E501
            return data

    def credential_service_delete_credential_with_http_info(self, cluster_id: 'str', id: 'str', **kwargs) -> 'V1DeleteCredentialResponse':  # noqa: E501
        """credential_service_delete_credential  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.credential_service_delete_credential_with_http_info(cluster_id, id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str id: (required)
        :return: V1DeleteCredentialResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['cluster_id', 'id']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method credential_service_delete_credential" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'cluster_id' is set
        if self.api_client.client_side_validation and ('cluster_id' not in params or
                                                       params['cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `cluster_id` when calling `credential_service_delete_credential`")  # noqa: E501
        # verify the required parameter 'id' is set
        if self.api_client.client_side_validation and ('id' not in params or
                                                       params['id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `id` when calling `credential_service_delete_credential`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'cluster_id' in params:
            path_params['clusterId'] = params['cluster_id']  # noqa: E501
        if 'id' in params:
            path_params['id'] = params['id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{clusterId}/credentials/{id}', 'DELETE',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1DeleteCredentialResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def credential_service_get_credential(self, cluster_id: 'str', id: 'str', **kwargs) -> 'V1Credential':  # noqa: E501
        """credential_service_get_credential  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.credential_service_get_credential(cluster_id, id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str id: (required)
        :return: V1Credential
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.credential_service_get_credential_with_http_info(cluster_id, id, **kwargs)  # noqa: E501
        else:
            (data) = self.credential_service_get_credential_with_http_info(cluster_id, id, **kwargs)  # noqa: E501
            return data

    def credential_service_get_credential_with_http_info(self, cluster_id: 'str', id: 'str', **kwargs) -> 'V1Credential':  # noqa: E501
        """credential_service_get_credential  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.credential_service_get_credential_with_http_info(cluster_id, id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str id: (required)
        :return: V1Credential
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['cluster_id', 'id']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method credential_service_get_credential" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'cluster_id' is set
        if self.api_client.client_side_validation and ('cluster_id' not in params or
                                                       params['cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `cluster_id` when calling `credential_service_get_credential`")  # noqa: E501
        # verify the required parameter 'id' is set
        if self.api_client.client_side_validation and ('id' not in params or
                                                       params['id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `id` when calling `credential_service_get_credential`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'cluster_id' in params:
            path_params['clusterId'] = params['cluster_id']  # noqa: E501
        if 'id' in params:
            path_params['id'] = params['id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{clusterId}/credentials/{id}', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1Credential',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def credential_service_list_credentials(self, cluster_id: 'str', **kwargs) -> 'V1ListCredentialsResponse':  # noqa: E501
        """credential_service_list_credentials  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.credential_service_list_credentials(cluster_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str page_token:
        :param str limit:
        :param list[str] user_ids:
        :param list[str] credential_type_in:
        :param list[str] credential_type_not_in:
        :return: V1ListCredentialsResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.credential_service_list_credentials_with_http_info(cluster_id, **kwargs)  # noqa: E501
        else:
            (data) = self.credential_service_list_credentials_with_http_info(cluster_id, **kwargs)  # noqa: E501
            return data

    def credential_service_list_credentials_with_http_info(self, cluster_id: 'str', **kwargs) -> 'V1ListCredentialsResponse':  # noqa: E501
        """credential_service_list_credentials  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.credential_service_list_credentials_with_http_info(cluster_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str page_token:
        :param str limit:
        :param list[str] user_ids:
        :param list[str] credential_type_in:
        :param list[str] credential_type_not_in:
        :return: V1ListCredentialsResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['cluster_id', 'page_token', 'limit', 'user_ids', 'credential_type_in', 'credential_type_not_in']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method credential_service_list_credentials" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'cluster_id' is set
        if self.api_client.client_side_validation and ('cluster_id' not in params or
                                                       params['cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `cluster_id` when calling `credential_service_list_credentials`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'cluster_id' in params:
            path_params['clusterId'] = params['cluster_id']  # noqa: E501

        query_params = []
        if 'page_token' in params:
            query_params.append(('pageToken', params['page_token']))  # noqa: E501
        if 'limit' in params:
            query_params.append(('limit', params['limit']))  # noqa: E501
        if 'user_ids' in params:
            query_params.append(('userIds', params['user_ids']))  # noqa: E501
            collection_formats['userIds'] = 'multi'  # noqa: E501
        if 'credential_type_in' in params:
            query_params.append(('credentialTypeIn', params['credential_type_in']))  # noqa: E501
            collection_formats['credentialTypeIn'] = 'multi'  # noqa: E501
        if 'credential_type_not_in' in params:
            query_params.append(('credentialTypeNotIn', params['credential_type_not_in']))  # noqa: E501
            collection_formats['credentialTypeNotIn'] = 'multi'  # noqa: E501

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{clusterId}/credentials', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1ListCredentialsResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def credential_service_update_credential(self, credential_cluster_id: 'str', credential_id: 'str', body: 'V1UpdateCredentialRequest', **kwargs) -> 'V1Credential':  # noqa: E501
        """credential_service_update_credential  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.credential_service_update_credential(credential_cluster_id, credential_id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str credential_cluster_id: (required)
        :param str credential_id: (required)
        :param V1UpdateCredentialRequest body: (required)
        :return: V1Credential
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.credential_service_update_credential_with_http_info(credential_cluster_id, credential_id, body, **kwargs)  # noqa: E501
        else:
            (data) = self.credential_service_update_credential_with_http_info(credential_cluster_id, credential_id, body, **kwargs)  # noqa: E501
            return data

    def credential_service_update_credential_with_http_info(self, credential_cluster_id: 'str', credential_id: 'str', body: 'V1UpdateCredentialRequest', **kwargs) -> 'V1Credential':  # noqa: E501
        """credential_service_update_credential  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.credential_service_update_credential_with_http_info(credential_cluster_id, credential_id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str credential_cluster_id: (required)
        :param str credential_id: (required)
        :param V1UpdateCredentialRequest body: (required)
        :return: V1Credential
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['credential_cluster_id', 'credential_id', 'body']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method credential_service_update_credential" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'credential_cluster_id' is set
        if self.api_client.client_side_validation and ('credential_cluster_id' not in params or
                                                       params['credential_cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `credential_cluster_id` when calling `credential_service_update_credential`")  # noqa: E501
        # verify the required parameter 'credential_id' is set
        if self.api_client.client_side_validation and ('credential_id' not in params or
                                                       params['credential_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `credential_id` when calling `credential_service_update_credential`")  # noqa: E501
        # verify the required parameter 'body' is set
        if self.api_client.client_side_validation and ('body' not in params or
                                                       params['body'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `body` when calling `credential_service_update_credential`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'credential_cluster_id' in params:
            path_params['credential.clusterId'] = params['credential_cluster_id']  # noqa: E501
        if 'credential_id' in params:
            path_params['credential.id'] = params['credential_id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{credential.clusterId}/credentials/{credential.id}', 'POST',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1Credential',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)



import typing
from dataclasses import dataclass, field


@dataclass
class ScratchSpace:
    """Represents compute node hard drives mounted in an experiment to be used as scratch space.

    Parameters
    ----------
    size_gb
        Size of the hard disk to mount in GB.
    mount_path
        Path to where the hard disk should be mounted.
    """
    size_gb: int = 100
    mount_path: str = '/scratch'


@dataclass
class Resources:
    """Represents compute node hard drives mounted in an experiment to be used as scratch space.

    Parameters
    ----------
    instance_type
        Instance type to use to launch the experiment
    use_spot
        If ``True``, inturuptable instance types will be used to launch the
        experiment computations. This trades the potential of longer
        times to completion for drastically reduced prices.
    cpus
        Number of CPUs to use per exeriment. If more CPUs are requested than
        exist on the compute node, the experiment will run with the total
        number of cpus which are available on a machine type. If zero (the
        default) then use all cpus available on a machine.
    gpus
        Number of GPUs to use per exeriment (requires an instance type
        to be selected which contains GPUs). If more GPUs are requested than
        exist on the compute node, the experiment will run with the total
        number of gpus which are available on a machine type. If zero (the
        default) then use all gpus available on a machine.
    storage_gb
        Size of the hard disk (GB) used to mount the root of the experiment
        file system. Must be >= 100 GB.
    extra
        Additional selectors to use to request machines. Not for general use
        unless instructed to by the grid team as part of your support contract.
    """
    instance_type: str = 't2.medium'
    use_spot: bool = False
    cpus: int = 0
    gpus: int = 0
    storage_gb: int = 100
    extra: typing.Dict[str, str] = field(default_factory=dict)


@dataclass
class Actions:
    """Actions define commands executed at specific points in the run lifecycle.

    Actions are able to influence experiment image generation
    before building source code, after building source code,
    before starting the user program, or after the user program ends.

    Parameters
    ----------
    on_build
        commands to add as RUN instructions. These instructions are
        executed before any dependencies are installed. Note: this
        cannot be set if using custom Dockerfile
    on_build_start
        commands passed to the image builder which are interpreted
        as RUN commands in a Dockerfile. Executes before installing
        dependencies from package manager.
    on_build_end
        commands passed to the image builder which are interpreted
        as RUN commands in a Dockerfile. Executes after installing
        dependencies from package manager.
    on_experiment_start
        allows users to specify commands that need to be executed
        sequentially before the main experiment process starts.
        This command will be executed on every experiment that the
        run creates.
    on_experiment_end
        allows users to specify commands that need to be executed
        sequentially after the main experiment process ends. This
        command will be executed on every experiment that the run
        creates.
    """
    on_build: typing.List[str] = None
    on_build_start: typing.List[str] = None
    on_build_end: typing.List[str] = None
    on_experiment_start: typing.List[str] = None
    on_experiment_end: typing.List[str] = None



import collections
import contextlib
from enum import Enum
from typing import Mapping, Optional, Type, TypeVar, Any

import click
import rich
from rich import box
from rich.console import Console, RenderableType
from rich.rule import Rule
from rich.style import Style

try:
    from rich.console import RenderGroup
except ImportError:
    # this was renamed to rich.console.Group as of Rich 11.0
    # see https://github.com/Textualize/rich/releases/tag/v11.0.0 for more details
    from rich.console import Group as RenderGroup
from rich.markup import escape
from rich.padding import Padding
from rich.panel import Panel
from rich.table import Table
from rich.theme import Theme

from grid.cli.utilities import is_latest_version
from grid.metadata import __package_name__
import grid.sdk.env as env


class GridColors(str, Enum):
    """These are the grid brand colors made slightly darker so they work with both black and white backgrounds."""

    GREEN = "#47ff6f"  # Original: #78FF96
    PERIWINKLE = "#70a5ff"  # Original: #A4C6FF
    ORANGE = "#e66700"  # Original: #FF7300
    TEAL = "#14857e"  # Original: #18A096
    PURPLE = "#5233ff"  # Original: #6950FF


GRID_THEME = Theme(
    {
        "title": "bold",
        "header": "bold",
        "warning": "bold red",
        "note": f"bold {GridColors.ORANGE}",
        "command_path": f"bold {GridColors.PURPLE}",
        "argument": GridColors.TEAL,
        "option": GridColors.ORANGE,
        "command": GridColors.TEAL,
        "maintenance": GridColors.GREEN,
    },
    inherit=False,
)


def grid_console():
    """Creates a ``rich.console.Console`` equipped with the GRID_THEME."""
    return Console(theme=GRID_THEME, force_terminal=True)


@contextlib.contextmanager
def console_from_formatter(formatter):
    """Context manager which substitutes a formatter for a ``grid_console`` and writes the output back to the
    formatter.
    """
    console = grid_console()
    with console.capture() as capture:
        yield console
    formatter.write(capture.get())


def render(renderable: RenderableType):
    """Renders a ``rich.RenderableType`` to a string and returns it."""
    console = grid_console()
    with console.capture() as capture:
        console.print(renderable)
    return capture.get()


def render_maintenance_window(renderable: RenderableType):
    """Utility method to render a ``rich.RenderableType`` alongside a 'Maintenance Window' header."""
    msg = render(
        RenderGroup(
            Rule("[bold]NOTICE", style=Style(color=GridColors.GREEN, bold=True)),
            '[maintenance]Upcoming Maintenance[/maintenance]:', pad(renderable),
            Rule(style=Style(color=GridColors.GREEN, bold=True))
        )
    )
    return msg


def render_warning(renderable: RenderableType):
    """Utility method to render a ``rich.RenderableType`` alongside a 'Warning' header."""
    return render(RenderGroup(
        '[warning]Warning[/warning]:',
        pad(renderable),
    ))


def render_error(renderable: RenderableType):
    """Utility method to render a ``rich.RenderableType`` alongside an 'Error' header."""
    return render(RenderGroup(
        '[error]Error[/error]:',
        pad(renderable),
    ))


def render_note(renderable: RenderableType):
    """Utility method to render a ``rich.RenderableType`` alongside a 'Note' header."""
    return render(RenderGroup(
        '[note]Note[/note]:',
        pad(renderable),
    ))


def pad(renderable: RenderableType):
    """Pads the given ``rich.RenderableType`` to the standard padding used in the Grid CLI."""
    padding = (0, 4, 1, 4)
    if isinstance(renderable, Table):
        padding = (0, 4, 1, 2)
    return Padding(renderable, padding)


COMMAND_TYPE = TypeVar("COMMAND_TYPE", bound=click.Command)


def make_rich_command(cls: Type[COMMAND_TYPE]) -> Type[COMMAND_TYPE]:
    """Creates and returns a subclass of the given ``click.Command`` type which renders help text with rich.

    Parameters
    ----------
    cls
        The ``click.Command`` subclass to enhance with rich.
    """
    class RichCommand(cls):
        def __init__(self, name, **kwargs):
            context_settings = kwargs.pop("context_settings", {})
            context_settings.update(
                {  # Let rich handle line wrapping
                    'terminal_width': 10000,
                    'max_content_width': 10000,
                }
            )
            super().__init__(name, context_settings=context_settings, **kwargs)

        def get_max_lhs_column_width(self, ctx) -> int:
            all_lens = []
            all_lens += [len(argument) for argument in self.get_arguments(ctx).keys()]
            all_lens += [len(option) for option in self.get_options(ctx).keys()]
            return max(all_lens)

        def format_help(self, ctx, formatter):
            self.format_help_text(ctx, formatter)
            self.format_usage(ctx, formatter)
            self.format_options(ctx, formatter)
            self.format_arguments(ctx, formatter)
            self.format_epilog(ctx, formatter)

        def format_help_text(self, ctx, formatter):
            with console_from_formatter(formatter) as console:
                if self.help is not None:
                    console.print(pad(Panel(self.help, style="title", expand=False)))

        def format_usage(self, ctx, formatter):
            with console_from_formatter(formatter) as console:
                self.print_header("Usage:", console)

                usage_pieces = []
                for usage_piece in self.collect_usage_pieces(ctx):
                    usage_pieces += list(usage_piece.split(" "))

                usage_pieces = [
                    f"[command_path]{ctx.command_path}[/command_path]", f"[option]{self.options_metavar}[/option]"
                ]
                for param in self.get_params(ctx):
                    if isinstance(param, click.Argument):
                        usage_pieces += [f"[argument]{param.get_usage_pieces(ctx)[0]}[/argument]"]

                subcommand_metavar = getattr(self, "subcommand_metavar", "")
                usage_pieces.append(f"[command]{subcommand_metavar}[/command]")

                console.print(pad(" ".join(usage_pieces)))

        @staticmethod
        def get_arguments(ctx):
            arguments = collections.OrderedDict()
            for p in ctx.command.params:
                if isinstance(p, click.Argument):
                    arguments[p.name.upper()] = getattr(p, 'help', '')
            return arguments

        def get_options(self, ctx):
            options = collections.OrderedDict()
            for param in self.params:
                help_record = param.get_help_record(ctx)
                if help_record is not None:
                    options[help_record[0]] = help_record[1]
            options["--help"] = "Get detailed info."
            return options

        @staticmethod
        def print_header(header, console):
            console.print(header, style="header")

        def get_help_table(self, entries: Mapping[str, str], ctx, style):
            table = Table(show_header=False, box=box.MINIMAL)
            table.add_column(justify="left", style=style, width=self.get_max_lhs_column_width(ctx))
            table.add_column(justify="left")
            for entry, help_text in entries.items():
                table.add_row(escape(entry), help_text)
            return table

        def format_arguments(self, ctx, formatter):
            with console_from_formatter(formatter) as console:
                arguments = self.get_arguments(ctx)
                if len(arguments) > 0:
                    self.print_header("Arguments:", console)
                    console.print(pad(self.get_help_table(arguments, ctx, "argument")))

        def format_options(self, ctx, formatter):
            with console_from_formatter(formatter) as console:
                options = self.get_options(ctx)
                if len(options) > 0:
                    self.print_header("Options:", console)
                    console.print(pad(self.get_help_table(options, ctx, "option")))

    return RichCommand


GROUP_TYPE = TypeVar("GROUP_TYPE", bound=click.Group)


def make_rich_group(cls: Type[GROUP_TYPE]) -> Type[GROUP_TYPE]:
    """Creates and returns a subclass of the given ``click.Group`` type which renders help text with rich.

    Parameters
    ----------
    cls
        The ``click.Group`` subclass to enhance with rich.
    """
    class RichGroup(make_rich_command(cls)):
        def command(self, *args, **kwargs):
            cls = kwargs.get("cls", click.Command)
            kwargs["cls"] = make_rich_command(cls)
            return super().command(*args, **kwargs)

        def group(self, *args, **kwargs):
            cls = kwargs.get("cls", click.Group)
            kwargs["cls"] = make_rich_group(cls)
            return super().command(*args, **kwargs)

        def get_max_lhs_column_width(self, ctx) -> int:
            all_lens = [super().get_max_lhs_column_width(ctx)]
            all_lens += [len(command) for command in self.get_commands(ctx).keys()]
            return max(all_lens)

        def format_help(self, ctx, formatter):
            self.format_help_text(ctx, formatter)
            self.format_usage(ctx, formatter)
            self.format_options(ctx, formatter)
            self.format_arguments(ctx, formatter)
            self.format_commands(ctx, formatter)
            self.format_learn_more(ctx, formatter)
            self.format_epilog(ctx, formatter)

        def get_commands(self, ctx):
            commands = collections.OrderedDict()
            for command in self.list_commands(ctx):
                command = self.commands.get(command)
                commands[command.name] = command.get_short_help_str(limit=120)
            return commands

        def format_commands(self, ctx, formatter):
            with console_from_formatter(formatter) as console:
                self.print_header("Commands:", console)
                console.print(pad(self.get_help_table(self.get_commands(ctx), ctx, "command")))

        def format_learn_more(self, ctx, formatter):
            with console_from_formatter(formatter) as console:
                self.print_header("Learn more:", console)
                console.print(
                    pad(
                        f"Use [command_path]{ctx.command_path.strip()}[/command_path] [command]COMMAND[/command] "
                        "[option]--help[/option] for more information about a command."
                    )
                )

    return RichGroup


def deprecate_option(workaround: str, ctx: click.Context, param: click.Parameter, value: Any):
    if value:
        Console().print(
            f"[bold red] WARNING: [/bold red] The command option `{param.opts[0]}` "
            f"has been deprecated and will be removed in a future release. \n "
            f"          Please use {workaround} instead."
        )
    return value


def make_rich(cls: Type[COMMAND_TYPE]) -> Type[COMMAND_TYPE]:
    """Utility function to enhance a ``click.Command`` or ``click.Group`` type with rich by choosing the correct
    ``make_rich_*`` method.

    Parameters
    ----------
    cls
        The ``click.Command`` subclass to enhance with rich.
    """
    if issubclass(cls, click.Group):
        return make_rich_group(cls)
    return make_rich_command(cls)


def command(name=None, cls=None, **attrs):
    """An override for the ``click.command`` decorator which adds rich help text."""
    if cls is None:
        cls = click.Command
    cls = make_rich(cls)
    return click.command(name=name, cls=cls, **attrs)


def group(name=None, cls=None, **attrs):
    """An override for the ``click.group`` decorator which adds rich help text."""
    if cls is None:
        cls = click.Group
    cls = make_rich(cls)
    return click.group(name=name, cls=cls, **attrs)


def main_group(name=None, cls=None, **attrs):
    if cls is None:
        cls = click.Group
    cls = make_rich(cls)

    class RichMainGroup(cls):
        def format_epilog(self, ctx, formatter):
            with console_from_formatter(formatter) as console:
                self.print_header("Examples:", console)
                console.print(
                    pad(
                        RenderGroup(
                            "[command_path]grid[/command_path] [command]login[/command]",
                            "[command_path]grid[/command_path] [command]version[/command]",
                        )
                    )
                )

                self.print_header("Feedback:", console)
                console.print(
                    pad(
                        "For questions or feedback Join our slack! "
                        "[underline][link=https://bit.ly/3fUOeDU]https://bit.ly/3fUOeDU[/link][/underline]"
                    )
                )

        def parse_args(self, ctx, args):
            # NOTE: Uncomment lines bellow and replace timestamp to print maintenance window message
            #       before every command is run (when a maintenance window / downtime is scheduled)
            # message = render_maintenance_window(
            #     "Grid will be unavailable on Sunday May 29, 2022 from 9pm-11pm EST. \n"
            #     "Please ensure you have no active Runs, Sessions, or Experiments.",
            # )
            # click.echo(message)

            if not is_latest_version():
                message = render_warning(
                    f"A newer version of [command_path]{__package_name__}[/command_path] is available. To upgrade, "
                    f"please run: pip install [command_path]{__package_name__}[/command_path] --upgrade"
                )
                if not env.IGNORE_WARNINGS:
                    click.echo(message)
            return super().parse_args(ctx, args)

    cls = RichMainGroup
    return click.group(name=name, cls=cls, **attrs)


def argument(*param_decls, **attrs):
    """An override for the ``click.argument`` decorator that adds a ``help`` argument, similar to ``click.Command``."""
    cls = attrs.get("cls", click.Argument)

    class HelpArgument(cls):
        """A ``click.Argument`` which includes a short help message."""
        def __init__(self, *args, **kwargs):
            self.help = kwargs.pop("help", None)
            super().__init__(*args, **kwargs)

    attrs["cls"] = HelpArgument
    return click.argument(*param_decls, **attrs)


def deprecate_grid_options(cls: Optional[Type[click.Command]] = None):
    """Creates a custom ``click.Command`` class that gives a warning if parameters beginning with ``grid_`` or ``g_``
    are used.

    Parameters
    ----------
    cls
        Optionally pass the ``click.Command`` class to extend.
    """
    cls = cls or click.Command

    class DeprecatedCommand(cls):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def format_epilog(self, ctx, formatter):
            with console_from_formatter(formatter) as console:
                if self.epilog:
                    console.print(self.epilog)
                console.print('[note]Note[/note]:')
                console.print(
                    pad(
                        'Parameters prefixed with "grid_" or "g_" are deprecated and will be removed in a later '
                        'release. Use their non-prefixed variants instead.'
                    )
                )

        def parse_args(self, ctx, args):
            new_args = []
            for arg in args:
                if str(arg).startswith('--grid_') or str(arg).startswith('--g_'):
                    preferred = arg.replace('--grid_', '--').replace('--g_', '--')
                    # If the preferred param actually exists (is in the opts for anything in self.params) then we give a
                    # deprecation warning and replace the arg. If not, we do nothing and click will give a 'does not
                    # exist' error message.
                    if any(preferred in param.opts for param in self.params):
                        message = render_warning(
                            f'The [option]{arg}[/option] parameter is deprecated and will be removed in a later '
                            f'release. Use [option]{preferred}[/option] instead.'
                        )
                        if not env.IGNORE_WARNINGS:
                            click.echo(message)
                        arg = preferred
                new_args.append(arg)
            return super().parse_args(ctx, new_args)

    return DeprecatedCommand


def deprecate_and_alias(aliases: Mapping[str, str], cls: Optional[Type[GROUP_TYPE]] = None) -> Type[GROUP_TYPE]:
    """Creates a custom ``click.Group`` class that swaps commands with their aliases and gives a deprecation warning if
    the aliased commands are used.
    Parameters
    ----------
    aliases
        The mapping from old command name to new command name (alias).
    cls
        Optionally pass the ``click.Command`` class to extend.
    """
    cls = cls or click.Group

    # Referencing from: https://stackoverflow.com/questions/46641928/python-click-multiple-command-names
    class AliasedGroup(cls):
        def get_command(self, ctx, cmd_name):
            if cmd_name in aliases:
                preferred = aliases[cmd_name]
                message = render_warning(
                    f'The [option]{cmd_name}[/option] command is deprecated and will be removed in a later release. '
                    f'Use [option]{preferred}[/option] instead.'
                )
                if not env.IGNORE_WARNINGS:
                    click.echo(message)
                cmd_name = preferred
            return super().get_command(ctx, cmd_name)

    return AliasedGroup



from typing import List

import click
import yaspin

from grid import Experiment, Run
from grid.cli import rich_click
from grid.sdk.rest.exceptions import GridException
from grid.sdk.sessions import Session


@rich_click.group()
def stop():
    """Stop Runs or Experiments."""
    pass


@stop.command()
@rich_click.argument('experiment_names', type=str, required=True, nargs=-1, help="The experiment names to stop.")
def experiment(experiment_names: List[str]):
    """Stop one or more EXPERIMENT_NAMES.

    This preserves progress completed up to this point, but stops
    further computations and any billing for the machines used.
    """
    for name in experiment_names:
        spinner = yaspin.yaspin(text=f'Stopping experiment {name}...', color="yellow")
        spinner.start()
        try:
            exp = Experiment(name=name)
            if not exp.exists:
                raise RuntimeError(
                    f"Experiment {name} does not exist in the cluster {env.CONTEXT}. "
                    f"If you are trying to cancel an experiment in another cluster, "
                    f"try setting the default cluster with "
                    f"`grid user set-default-cluster <cluster_name>` first"
                )
            if not exp.cancel():
                raise RuntimeError(f"Experiment {name} could not be stopped.")
            spinner.ok("✔")
            styled_name = click.style(name, fg='blue')
            click.echo(f'Experiment {styled_name} was stopped successfully.')
        except GridException as e:
            if spinner:
                spinner.fail("✘")
            raise click.ClickException(str(e))
        except Exception as e:
            if spinner:
                spinner.fail("✘")
            raise click.ClickException(f"Stopping failed for experiment {name}: {e}")
        finally:
            spinner.stop()


@stop.command()
@rich_click.argument('run_names', type=str, required=True, nargs=-1, help="The run names to stop.")
def run(run_names: List[str]):
    """Stop one or more RUN_NAMES.

    This preserves progress completed up to this point, but stops
    further computations and any billing for the machines used.
    """
    for name in run_names:
        spinner = yaspin.yaspin(text=f'Stopping run {name}...', color="yellow")
        spinner.start()
        try:
            run_obj = Run(name=name)
            if not run_obj.exists:
                raise RuntimeError(
                    f"Run {name} does not exist in the cluster {env.CONTEXT}. "
                    f"If you are trying to cancel a run in another cluster, "
                    f"try setting the default cluster with "
                    f"`grid user set-default-cluster <cluster_name>` first"
                )
            if not run_obj.cancel():
                raise RuntimeError(f"Run {name} could not be stopped.")
            spinner.ok("✔")
            styled_name = click.style(name, fg='blue')
            click.echo(f'Run {styled_name} was stopped successfully.')
        except GridException as e:
            if spinner:
                spinner.fail("✘")
            raise click.ClickException(str(e))
        except Exception as e:
            if spinner:
                spinner.fail("✘")
            raise click.ClickException(f"Stopping failed for run {name}: {e}")
        finally:
            spinner.stop()
